__all__ = [
    "SpeakerBaseModel",
    "SpeakerMimic",
    "SpeakerPico",
    "SpeakerGttsMPV"
]

from epy_reader.speakers.base import SpeakerBaseModel
from epy_reader.speakers.mimic import SpeakerMimic
from epy_reader.speakers.pico import SpeakerPico
from epy_reader.speakers.gtts_mpv import SpeakerGttsMPV
